/*
 * kgraft_patch_bsc984764
 *
 * Fix for CVE-2016-4470, bsc#984764
 *
 *  Upstream commit:
 *  38327424b40b ("KEYS: potential uninitialized variable")
 *
 *  SLE12 commit:
 *  87cee7ed363db7a0871c1642665abd5047cf77fb
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/key.h>

#include "local.h"

static int (*kgr__key_link_begin)(struct key *keyring, const struct keyring_index_key *index_key, struct assoc_array_edit **_edit);
static struct mutex *kgr_key_construction_mutex;
static void (*kgr_key_schedule_gc)(time_t gc_at);
static void (*kgr__key_link)(struct key *key, struct assoc_array_edit **_edit);
static void (*kgr__key_link_end)(struct key *keyring, const struct keyring_index_key *index_key, struct assoc_array_edit *edit);
static unsigned *kgr_key_gc_delay;

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "__key_link_begin", (void *)&kgr__key_link_begin },
	{ "key_construction_mutex", (void *)&kgr_key_construction_mutex },
	{ "key_schedule_gc", (void *)&kgr_key_schedule_gc },
	{ "__key_link", (void *)&kgr__key_link },
	{ "__key_link_end", (void *)&kgr__key_link_end },
	{ "key_gc_delay", (void *)&kgr_key_gc_delay },
};

/* patched function */
int kgr_key_reject_and_link(struct key *key, unsigned timeout, unsigned error,
	struct key *keyring, struct key *authkey)
{
	struct assoc_array_edit *edit;
	struct timespec now;
	int ret, awaken, link_ret = 0;

	key_check(key);
	key_check(keyring);

	awaken = 0;
	ret = -EBUSY;

	if (keyring)
		link_ret = kgr__key_link_begin(keyring, &key->index_key, &edit);

	mutex_lock(kgr_key_construction_mutex);

	/* can't instantiate twice */
	if (!test_bit(KEY_FLAG_INSTANTIATED, &key->flags)) {
		/* mark the key as being negatively instantiated */
		atomic_inc(&key->user->nikeys);
		set_bit(KEY_FLAG_NEGATIVE, &key->flags);
		set_bit(KEY_FLAG_INSTANTIATED, &key->flags);
		key->type_data.reject_error = -error;
		now = current_kernel_time();
		key->expiry = now.tv_sec + timeout;
		kgr_key_schedule_gc(key->expiry + *kgr_key_gc_delay);

		if (test_and_clear_bit(KEY_FLAG_USER_CONSTRUCT, &key->flags))
			awaken = 1;

		ret = 0;

		/* and link it into the destination keyring */
		if (keyring && link_ret == 0)
			kgr__key_link(key, &edit);

		/* disable the authorisation key */
		if (authkey)
			key_revoke(authkey);
	}

	mutex_unlock(kgr_key_construction_mutex);

	/*
	 * Fix CVE-2016-4470
	 *  -if (keyring)
	 *  +next line
	 */
	if (keyring && link_ret == 0)
		kgr__key_link_end(keyring, &key->index_key, edit);

	/* wake up anyone waiting for a key to be constructed */
	if (awaken)
		wake_up_bit(&key->flags, KEY_FLAG_USER_CONSTRUCT);

	return ret == 0 ? link_ret : ret;
}

static int kgr_patch_bsc984764_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc984764_init(void)
{
	return kgr_patch_bsc984764_kallsyms();
}
